@@ -324,6 +324,394 @@ START_TEST(test_match_prefix)
 END_TEST
 
 
+START_TEST(test_match_prefix_strlen)
+{
+	/* Copyright (c) 2022 the CivetWeb developers */
+	ck_assert_int_eq(5, match_prefix_strlen("/Test", "/test"));
+	ck_assert_int_eq(-1, match_prefix_strlen("/Test", "/my/test"));
+	ck_assert_int_eq(3, match_prefix_strlen("/my", "/my/test"));
+	ck_assert_int_eq(-1, match_prefix_strlen("/my$", "/my/test"));
+	ck_assert_int_eq(8, match_prefix_strlen("/*/Test", "/my/test"));
+
+	ck_assert_int_eq(17,
+	                 match_prefix_strlen("/api/*/*.cgi", "/api/obj/prop.cgi"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/abc/*/*.cgi", "/api/obj/prop.cgi"));
+	ck_assert_int_eq(18,
+	                 match_prefix_strlen("/api/*/*.cgi", "/api/obj/other.cgi"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/api/*/*.cgi",
+	                                     "/api/obj/too/deep.cgi"));
+	ck_assert_int_eq(17,
+	                 match_prefix_strlen("/api/*/*.cgi$", "/api/obj/prop.cgi"));
+	ck_assert_int_eq(18,
+	                 match_prefix_strlen("/api/*/*.cgi$",
+	                                     "/api/obj/other.cgi"));
+	ck_assert_int_eq(17,
+	                 match_prefix_strlen("/api/*/*.cgi",
+	                                     "/api/obj/prop.cgiZZZ"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/api/*/*.cgi$",
+	                                     "/api/obj/prop.cgiZZZ"));
+	ck_assert_int_eq(-1, match_prefix_strlen("/*/*.cgi", "/api/obj/prop.cgi"));
+
+	ck_assert_int_eq(7, match_prefix_strlen("I????IT", "ItestIT"));
+	ck_assert_int_eq(-1, match_prefix_strlen("I????IT", "IseeIT"));
+	ck_assert_int_eq(23, match_prefix_strlen("**$", "EveryThing/matches this"));
+	ck_assert_int_eq(23,
+	                 match_prefix_strlen("?**$", "EveryThing/matches this"));
+	ck_assert_int_eq(23,
+	                 match_prefix_strlen("**?$", "EveryThing/matches this"));
+	ck_assert_int_eq(0, match_prefix_strlen("**$", ""));
+	ck_assert_int_eq(-1, match_prefix_strlen("?**$", ""));
+	ck_assert_int_eq(-1, match_prefix_strlen("**?$", ""));
+	ck_assert_int_eq(-1, match_prefix_strlen("/**?$", "/"));
+	ck_assert_int_eq(1, match_prefix_strlen("/**$", "/"));
+	ck_assert_int_eq(-1, match_prefix_strlen("//", "/"));
+	ck_assert_int_eq(1, match_prefix_strlen("/", "//"));
+	ck_assert_int_eq(-1, match_prefix_strlen("/$", "//"));
+
+	/* ? pattern should not match / character */
+	ck_assert_int_eq(-1, match_prefix_strlen("/?", "//"));
+	ck_assert_int_eq(3, match_prefix_strlen("/?/$", "/a/"));
+	ck_assert_int_eq(-1, match_prefix_strlen("/?/$", "///"));
+
+	/* Pattern From UserManual.md */
+	ck_assert_int_eq(20,
+	                 match_prefix_strlen("**.cgi$", "anywhere/anyname.cgi"));
+	ck_assert_int_eq(-1, match_prefix_strlen("**.cgi$", "name.cgi.not.at.end"));
+	ck_assert_int_eq(4, match_prefix_strlen("/foo", "/foo"));
+	ck_assert_int_eq(4, match_prefix_strlen("/foo", "/foobar"));
+	ck_assert_int_eq(-1, match_prefix_strlen("/foo", "not.at.start./foo"));
+	ck_assert_int_eq(1, match_prefix_strlen("**a$|**b$", "a"));
+	ck_assert_int_eq(2, match_prefix_strlen("**a$|**b$", "xb"));
+	ck_assert_int_eq(-1, match_prefix_strlen("**a$|**b$", "abc"));
+
+	ck_assert_int_eq(14,
+	                 match_prefix_strlen("/data/????.css$", "/data/12.4.css"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/data/????.css$", "/data/12/4.css"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/data/????.css$", "/data/../4.css"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/data/????.css$", "/else/12.4.css"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/data/????.css$", "/data/1234.cssx"));
+	ck_assert_int_eq(13, match_prefix_strlen("/data/*.js$", "/data/1234.js"));
+	ck_assert_int_eq(17,
+	                 match_prefix_strlen("/data/*.js$", "/data/12345678.js"));
+	ck_assert_int_eq(-1, match_prefix_strlen("/data/*.js$", "/else/1234.js"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/data/*.js$", "/data/../some.js"));
+	ck_assert_int_eq(-1, match_prefix_strlen("/data/*.js$", "/data//x.js"));
+	ck_assert_int_eq(-1, match_prefix_strlen("/data/*.js$", "/data/./x.js"));
+	ck_assert_int_eq(34,
+	                 match_prefix_strlen("/api/*/*.cgi$",
+	                                     "/api/resourcetype/resourcename.cgi"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/api/*/*.cgi$",
+	                                     "/api/resourcename.cgi"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/*.jpg$|/*.jpeg$",
+	                                     "/somewhere/something.txt"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/*.jpg$|/*.jpeg$", "/something.txt"));
+	ck_assert_int_eq(10, match_prefix_strlen("/*.jpg$|/*.jpeg$", "/image.jpg"));
+	ck_assert_int_eq(11,
+	                 match_prefix_strlen("/*.jpg$|/*.jpeg$", "/image.jpeg"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/*.jpg$|/*.jpeg$",
+	                                     "/image.jpeg.exe"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/*.jpg$|/*.jpeg$", "/sub/image.jpg"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("/*.jpg$|/*.jpeg$",
+	                                     "/sub/image.jpeg"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("**.jpg$|**.jpeg$",
+	                                     "/somewhere/something.txt"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("**.jpg$|**.jpeg$", "/something.txt"));
+	ck_assert_int_eq(10, match_prefix_strlen("**.jpg$|**.jpeg$", "/image.jpg"));
+	ck_assert_int_eq(11,
+	                 match_prefix_strlen("**.jpg$|**.jpeg$", "/image.jpeg"));
+	ck_assert_int_eq(-1,
+	                 match_prefix_strlen("**.jpg$|**.jpeg$",
+	                                     "/image.jpeg.exe"));
+	ck_assert_int_eq(14,
+	                 match_prefix_strlen("**.jpg$|**.jpeg$", "/sub/image.jpg"));
+	ck_assert_int_eq(15,
+	                 match_prefix_strlen("**.jpg$|**.jpeg$",
+	                                     "/sub/image.jpeg"));
+	ck_assert_int_eq(14,
+	                 match_prefix_strlen("/*/*.jpg$|/*/*.jpeg$",
+	                                     "/sub/image.jpg"));
+	ck_assert_int_eq(15,
+	                 match_prefix_strlen("/*/*.jpg$|/*/*.jpeg$",
+	                                     "/sub/image.jpeg"));
+}
+END_TEST
+
+
+START_TEST(test_match_prefix_fuzz)
+{
+	/* Copyright (c) 2022 the CivetWeb developers */
+	{
+		/* From fuzz test */
+		const char *pat = "**cacc//d/?dad?";
+		const char *str =
+		    "dbbddb/cb/ddcdbcbbab/dcdcbbbcaaacdbcac/dbdcadaa/bcaca/d/a/adcad";
+		ck_assert_int_eq(6, match_prefix(pat, 1, str));
+		ck_assert_int_eq(63, match_prefix(pat, 2, str));
+		ck_assert_int_eq(61, match_prefix(pat, 3, str));
+		ck_assert_int_eq(62, match_prefix(pat, 4, str));
+		ck_assert_int_eq(52, match_prefix(pat, 5, str));
+		ck_assert_int_eq(-1, match_prefix(pat, 6, str));
+	}
+	{
+		/* From fuzz test */
+		const char *pat = "a/a*d**/?*/*cdd";
+		const char *str =
+		    "a/aaddba/ddadbaacac//bcaadbc/badaccbdadadcbb//ccd/dcbacdcddc//c";
+		ck_assert_int_eq(0, match_prefix(pat, 0, str));
+		ck_assert_int_eq(1, match_prefix(pat, 1, str));
+		ck_assert_int_eq(2, match_prefix(pat, 2, str));
+		ck_assert_int_eq(3, match_prefix(pat, 3, str));
+		ck_assert_int_eq(8, match_prefix(pat, 4, str));
+		ck_assert_int_eq(6, match_prefix(pat, 5, str));
+		ck_assert_int_eq(8, match_prefix(pat, 6, str));
+		ck_assert_int_eq(63, match_prefix(pat, 7, str));
+		ck_assert_int_eq(62, match_prefix(pat, 8, str));
+		ck_assert_int_eq(63, match_prefix(pat, 9, str));
+		ck_assert_int_eq(63, match_prefix(pat, 10, str));
+		ck_assert_int_eq(61, match_prefix(pat, 11, str));
+		ck_assert_int_eq(61, match_prefix(pat, 12, str));
+		ck_assert_int_eq(60, match_prefix(pat, 13, str));
+		ck_assert_int_eq(58, match_prefix(pat, 14, str));
+		ck_assert_int_eq(59, match_prefix(pat, 15, str));
+		ck_assert_int_eq(59, match_prefix_strlen(pat, str));
+	}
+	{
+		/* From fuzz test */
+		const char *pat = "cc/?**ba????**b";
+		const char *str =
+		    "cc/babdb/cbb/baa/da/cd///ccabbcdcdaa/dbacbdbadaccb/dbdcc/cdbbac";
+		ck_assert_int_eq(0, match_prefix(pat, 0, str));
+		ck_assert_int_eq(1, match_prefix(pat, 1, str));
+		ck_assert_int_eq(2, match_prefix(pat, 2, str));
+		ck_assert_int_eq(3, match_prefix(pat, 3, str));
+		ck_assert_int_eq(4, match_prefix(pat, 4, str));
+		ck_assert_int_eq(8, match_prefix(pat, 5, str));
+		ck_assert_int_eq(63, match_prefix(pat, 6, str));
+		ck_assert_int_eq(61, match_prefix(pat, 7, str));
+		ck_assert_int_eq(62, match_prefix(pat, 8, str));
+		ck_assert_int_eq(63, match_prefix(pat, 9, str));
+		ck_assert_int_eq(47, match_prefix(pat, 10, str));
+		ck_assert_int_eq(48, match_prefix(pat, 11, str));
+		ck_assert_int_eq(49, match_prefix(pat, 12, str));
+		ck_assert_int_eq(50, match_prefix(pat, 13, str));
+		ck_assert_int_eq(63, match_prefix(pat, 14, str));
+		ck_assert_int_eq(61, match_prefix(pat, 15, str));
+		ck_assert_int_eq(61, match_prefix_strlen(pat, str));
+	}
+	{
+		/* From fuzz test */
+		const char *pat = "?**ba*db*b?*/a?";
+		const char *str =
+		    "bd/bcdddabbd//bcb//acbcaaac/dcbbcdadabadba/bd/baadbabcc/a/bcb//";
+		ck_assert_int_eq(0, match_prefix(pat, 0, str));
+		ck_assert_int_eq(1, match_prefix(pat, 1, str));
+		ck_assert_int_eq(2, match_prefix(pat, 2, str));
+		ck_assert_int_eq(63, match_prefix(pat, 3, str));
+		ck_assert_int_eq(61, match_prefix(pat, 4, str));
+		ck_assert_int_eq(52, match_prefix(pat, 5, str));
+		ck_assert_int_eq(55, match_prefix(pat, 6, str));
+		ck_assert_int_eq(50, match_prefix(pat, 7, str));
+		ck_assert_int_eq(51, match_prefix(pat, 8, str));
+		ck_assert_int_eq(55, match_prefix(pat, 9, str));
+		ck_assert_int_eq(53, match_prefix(pat, 10, str));
+		ck_assert_int_eq(54, match_prefix(pat, 11, str));
+		ck_assert_int_eq(55, match_prefix(pat, 12, str));
+		ck_assert_int_eq(56, match_prefix(pat, 13, str));
+		ck_assert_int_eq(57, match_prefix(pat, 14, str));
+		ck_assert_int_eq(-1, match_prefix(pat, 15, str));
+		ck_assert_int_eq(-1, match_prefix_strlen(pat, str));
+	}
+	{
+		/* From fuzz test */
+		const char *pat = "?b*da*bc?c**//*";
+		const char *str =
+		    "dbadabcbcbbba/a///d//dcdd////daccbcaaa/a/bacddab/bdcbbdd/bbaa/a";
+		ck_assert_int_eq(0, match_prefix(pat, 0, str));
+		ck_assert_int_eq(1, match_prefix(pat, 1, str));
+		ck_assert_int_eq(2, match_prefix(pat, 2, str));
+		ck_assert_int_eq(13, match_prefix(pat, 3, str));
+		ck_assert_int_eq(4, match_prefix(pat, 4, str));
+		ck_assert_int_eq(5, match_prefix(pat, 5, str));
+		ck_assert_int_eq(13, match_prefix(pat, 6, str));
+		ck_assert_int_eq(12, match_prefix(pat, 7, str));
+		ck_assert_int_eq(9, match_prefix(pat, 8, str));
+		ck_assert_int_eq(10, match_prefix(pat, 9, str));
+		ck_assert_int_eq(9, match_prefix(pat, 10, str));
+		ck_assert_int_eq(13, match_prefix(pat, 11, str));
+		ck_assert_int_eq(63, match_prefix(pat, 12, str));
+		ck_assert_int_eq(62, match_prefix(pat, 13, str));
+		ck_assert_int_eq(29, match_prefix(pat, 14, str));
+		ck_assert_int_eq(38, match_prefix(pat, 15, str));
+		ck_assert_int_eq(38, match_prefix_strlen(pat, str));
+	}
+	{
+		/* From fuzz test */
+		const char *pat = "?**c**dc**b*c*a";
+		const char *str = "cba/a/bbcdcccc/bcacacdadab/dad/";
+		ck_assert_int_eq(0, match_prefix(pat, 0, str));
+		ck_assert_int_eq(1, match_prefix(pat, 1, str));
+		ck_assert_int_eq(3, match_prefix(pat, 2, str));
+		ck_assert_int_eq(31, match_prefix(pat, 3, str));
+		ck_assert_int_eq(21, match_prefix(pat, 4, str));
+		ck_assert_int_eq(26, match_prefix(pat, 5, str));
+		ck_assert_int_eq(31, match_prefix(pat, 6, str));
+		ck_assert_int_eq(30, match_prefix(pat, 7, str));
+		ck_assert_int_eq(11, match_prefix(pat, 8, str));
+		ck_assert_int_eq(14, match_prefix(pat, 9, str));
+		ck_assert_int_eq(31, match_prefix(pat, 10, str));
+		ck_assert_int_eq(26, match_prefix(pat, 11, str));
+		ck_assert_int_eq(26, match_prefix(pat, 12, str));
+		ck_assert_int_eq(21, match_prefix(pat, 13, str));
+		ck_assert_int_eq(26, match_prefix(pat, 14, str));
+		ck_assert_int_eq(25, match_prefix(pat, 15, str));
+		ck_assert_int_eq(25, match_prefix_strlen(pat, str));
+	}
+	{
+		/* From fuzz test */
+		const char *pat = "$|*ca|**c/c*b|a";
+		const char *str = "bcdaa/a//acdc/bac/caacacdcbcdc/";
+		ck_assert_int_eq(0, match_prefix(pat, 0, str));
+		ck_assert_int_eq(-1, match_prefix(pat, 1, str));
+		ck_assert_int_eq(0, match_prefix(pat, 2, str));
+		ck_assert_int_eq(5, match_prefix(pat, 3, str));
+		ck_assert_int_eq(2, match_prefix(pat, 4, str));
+		ck_assert_int_eq(-1, match_prefix(pat, 5, str));
+		ck_assert_int_eq(0, match_prefix(pat, 6, str));
+		ck_assert_int_eq(5, match_prefix(pat, 7, str));
+		ck_assert_int_eq(31, match_prefix(pat, 8, str));
+		ck_assert_int_eq(30, match_prefix(pat, 9, str));
+		ck_assert_int_eq(31, match_prefix(pat, 10, str));
+		ck_assert_int_eq(19, match_prefix(pat, 11, str));
+		ck_assert_int_eq(30, match_prefix(pat, 12, str));
+		ck_assert_int_eq(27, match_prefix(pat, 13, str));
+		ck_assert_int_eq(27, match_prefix(pat, 14, str));
+		ck_assert_int_eq(27, match_prefix(pat, 15, str));
+		ck_assert_int_eq(27, match_prefix_strlen(pat, str));
+	}
+	{
+		/* From fuzz test */
+		const char *pat = "*b?c|?$|*b?b**?";
+		const char *str = "cbbbddddda/dbcdbbbccdcb/a//ddab";
+		ck_assert_int_eq(0, match_prefix(pat, 0, str));
+		ck_assert_int_eq(10, match_prefix(pat, 1, str));
+		ck_assert_int_eq(4, match_prefix(pat, 2, str));
+		ck_assert_int_eq(5, match_prefix(pat, 3, str));
+		ck_assert_int_eq(-1, match_prefix(pat, 4, str));
+		ck_assert_int_eq(0, match_prefix(pat, 5, str));
+		ck_assert_int_eq(1, match_prefix(pat, 6, str));
+		ck_assert_int_eq(-1, match_prefix(pat, 7, str));
+		ck_assert_int_eq(0, match_prefix(pat, 8, str));
+		ck_assert_int_eq(10, match_prefix(pat, 9, str));
+		ck_assert_int_eq(4, match_prefix(pat, 10, str));
+		ck_assert_int_eq(5, match_prefix(pat, 11, str));
+		ck_assert_int_eq(4, match_prefix(pat, 12, str));
+		ck_assert_int_eq(10, match_prefix(pat, 13, str));
+		ck_assert_int_eq(31, match_prefix(pat, 14, str));
+		ck_assert_int_eq(31, match_prefix(pat, 15, str));
+		ck_assert_int_eq(31, match_prefix_strlen(pat, str));
+	}
+	{
+		/* From fuzz test */
+		const char *pat = "c|b?*$|*ca*/*ba";
+		const char *str = "bdccacb/aaadbadd/ccaca/c/cdcb/d";
+		ck_assert_int_eq(0, match_prefix(pat, 0, str));
+		ck_assert_int_eq(-1, match_prefix(pat, 1, str));
+		ck_assert_int_eq(0, match_prefix(pat, 2, str));
+		ck_assert_int_eq(1, match_prefix(pat, 3, str));
+		ck_assert_int_eq(2, match_prefix(pat, 4, str));
+		ck_assert_int_eq(7, match_prefix(pat, 5, str));
+		ck_assert_int_eq(-1, match_prefix(pat, 6, str));
+		ck_assert_int_eq(0, match_prefix(pat, 7, str));
+		ck_assert_int_eq(7, match_prefix(pat, 8, str));
+		ck_assert_int_eq(6, match_prefix(pat, 9, str));
+		ck_assert_int_eq(5, match_prefix(pat, 10, str));
+		ck_assert_int_eq(7, match_prefix(pat, 11, str));
+		ck_assert_int_eq(8, match_prefix(pat, 12, str));
+		ck_assert_int_eq(16, match_prefix(pat, 13, str));
+		ck_assert_int_eq(13, match_prefix(pat, 14, str));
+		ck_assert_int_eq(14, match_prefix(pat, 15, str));
+		ck_assert_int_eq(14, match_prefix_strlen(pat, str));
+	}
+}
+END_TEST
+
+
+START_TEST(test_mg_match)
+{
+	/* Copyright (c) 2022 the CivetWeb developers */
+	struct mg_match_context mcx;
+
+	ck_assert_int_eq(4, mg_match_alternatives("a*D", 3, "abcde", NULL));
+
+	memset(&mcx, 0, sizeof(mcx));
+	mcx.case_sensitive = 0;
+	ck_assert_int_eq(4, mg_match_alternatives("a*D", 3, "abcde", &mcx));
+	ck_assert_int_eq(1, (int)mcx.num_matches);
+	ck_assert_int_eq(2, (int)mcx.match[0].len);
+	ck_assert(!memcmp(mcx.match[0].str, "bc", 2));
+
+	memset(&mcx, 0, sizeof(mcx));
+	mcx.case_sensitive = 1;
+	ck_assert_int_eq(-1, mg_match_alternatives("a*D", 3, "abcde", &mcx));
+	ck_assert_int_eq(0, (int)mcx.num_matches);
+
+	memset(&mcx, 0, sizeof(mcx));
+	mcx.case_sensitive = 1;
+	ck_assert_int_eq(4, mg_match_alternatives("a??d", 4, "abcde", &mcx));
+	ck_assert_int_eq(1, (int)mcx.num_matches);
+	ck_assert_int_eq(2, (int)mcx.match[0].len);
+	ck_assert(!memcmp(mcx.match[0].str, "bc", 2));
+
+	memset(&mcx, 0, sizeof(mcx));
+	mcx.case_sensitive = 1;
+	ck_assert_int_eq(5, mg_match_alternatives("a??d*", 5, "abcde", &mcx));
+	ck_assert_int_eq(2, (int)mcx.num_matches);
+	ck_assert_int_eq(2, (int)mcx.match[0].len);
+	ck_assert(!memcmp(mcx.match[0].str, "bc", 2));
+	ck_assert_int_eq(1, (int)mcx.match[1].len);
+	ck_assert(!memcmp(mcx.match[1].str, "e", 1));
+
+	memset(&mcx, 0, sizeof(mcx));
+	mcx.case_sensitive = 1;
+	ck_assert_int_eq(4, mg_match_alternatives("a??d*", 5, "abcd", &mcx));
+	ck_assert_int_eq(2, (int)mcx.num_matches);
+	ck_assert_int_eq(2, (int)mcx.match[0].len);
+	ck_assert(!memcmp(mcx.match[0].str, "bc", 2));
+	ck_assert_int_eq(0, (int)mcx.match[1].len);
+
+	memset(&mcx, 0, sizeof(mcx));
+	mcx.case_sensitive = 0;
+	ck_assert_int_eq(2, mg_match_alternatives("a?|?B", 5, "ABC", &mcx));
+	ck_assert_int_eq(1, (int)mcx.num_matches);
+	ck_assert_int_eq(1, (int)mcx.match[0].len);
+	ck_assert(!memcmp(mcx.match[0].str, "B", 1));
+
+	memset(&mcx, 0, sizeof(mcx));
+	mcx.case_sensitive = 1;
+	ck_assert_int_eq(2, mg_match_alternatives("a?|?B", 5, "ABC", &mcx));
+	ck_assert_int_eq(1, (int)mcx.num_matches);
+	ck_assert_int_eq(1, (int)mcx.match[0].len);
+	ck_assert(!memcmp(mcx.match[0].str, "A", 1));
+}
+END_TEST
+
+
 START_TEST(test_remove_dot_segments)
 {
 	int i;
@@ -399,7 +787,7 @@ END_TEST
 
 START_TEST(test_is_valid_uri)
 {
-	/* is_valid_uri is superseeded by get_uri_type */
+	/* is_valid_uri is superseded by get_uri_type */
 	ck_assert_int_eq(2, get_uri_type("/api"));
 	ck_assert_int_eq(2, get_uri_type("/api/"));
 	ck_assert_int_eq(2,
@@ -464,7 +852,7 @@ END_TEST
 
 
 static int
-alloc_printf(char **buf, size_t size, const char *fmt, ...)
+alloc_vprintf_wrapper(char **buf, size_t size, const char *fmt, ...)
 {
 	/* Test helper function - adapted from unit_test.c */
 	/* Copyright (c) 2013-2015 the Civetweb developers */
@@ -482,23 +870,6 @@ alloc_printf(char **buf, size_t size, const char *fmt, ...)
 }
 
 
-static int
-alloc_printf2(char **buf, const char *fmt, ...)
-{
-	/* Test alternative implementation */
-	va_list ap;
-	int ret = 0;
-
-	mark_point();
-
-	va_start(ap, fmt);
-	ret = alloc_vprintf2(buf, fmt, ap);
-	va_end(ap);
-
-	return ret;
-}
-
-
 START_TEST(test_alloc_vprintf)
 {
 	/* Adapted from unit_test.c */
@@ -507,23 +878,23 @@ START_TEST(test_alloc_vprintf)
 	char buf[MG_BUF_LEN], *p = buf;
 	mark_point();
 
-	ck_assert(alloc_printf(&p, sizeof(buf), "%s", "hi") == 2);
+	ck_assert(alloc_vprintf_wrapper(&p, sizeof(buf), "%s", "hi") == 2);
 	ck_assert(p == buf);
 
-	ck_assert(alloc_printf(&p, sizeof(buf), "%s", "") == 0);
+	ck_assert(alloc_vprintf_wrapper(&p, sizeof(buf), "%s", "") == 0);
 	ck_assert(p == buf);
 
-	ck_assert(alloc_printf(&p, sizeof(buf), "") == 0);
+	ck_assert(alloc_vprintf_wrapper(&p, sizeof(buf), "") == 0);
 	ck_assert(p == buf);
 
 	/* Pass small buffer, make sure alloc_printf allocates */
-	ck_assert(alloc_printf(&p, 1, "%s", "hello") == 5);
+	ck_assert(alloc_vprintf_wrapper(&p, 1, "%s", "hello") == 5);
 	ck_assert(p != buf);
 	mg_free(p);
 	p = buf;
 
-	/* Test alternative implementation */
-	ck_assert(alloc_printf2(&p, "%s", "hello") == 5);
+	/* Test new wrapper implementation */
+	ck_assert(alloc_printf(&p, "%s", "hello") == 5);
 	ck_assert(p != buf);
 	mg_free(p);
 	p = buf;
@@ -698,15 +1069,277 @@ START_TEST(test_parse_port_string)
 	vec.ptr = "localhost:123";
 	vec.len = strlen(vec.ptr);
 	ret = parse_port_string(&vec, &so, &ip_family);
-	ck_assert_int_eq(ret, 1);
+	if (ret != 1) {
+		ck_abort_msg("IP of localhost seems to be unknown on this system (%i)",
+		             (int)ret);
+	}
 	if ((ip_family != 4) && (ip_family != 6)) {
-		ck_abort_msg("IP family must be 4 or 6 but is %i", (int)ip_family);
+		ck_abort_msg("IP family for localhost must be 4 or 6 but is %i",
+		             (int)ip_family);
 	}
 	ck_assert_int_eq((int)htons(so.lsa.sin.sin_port), (int)123);
 }
 END_TEST
 
 
+START_TEST(test_parse_http_headers)
+{
+	char buf[2048];
+	char *ptr;
+	int ret;
+	struct mg_header hdr[MG_MAX_HEADERS];
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 0);
+	ck_assert_ptr_eq(ptr, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].name, NULL);
+	ck_assert_ptr_eq(hdr[0].value, NULL);
+	ck_assert_ptr_eq(hdr[1].name, NULL);
+	ck_assert_ptr_eq(hdr[1].value, NULL);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "\r\n");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 0);
+	ck_assert_ptr_eq(ptr, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].name, NULL);
+	ck_assert_ptr_eq(hdr[0].value, NULL);
+	ck_assert_ptr_eq(hdr[1].name, NULL);
+	ck_assert_ptr_eq(hdr[1].value, NULL);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a\r\n");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, -1);
+	ck_assert_ptr_eq(ptr, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].name, NULL);
+	ck_assert_ptr_eq(hdr[0].value, NULL);
+	ck_assert_ptr_eq(hdr[1].name, NULL);
+	ck_assert_ptr_eq(hdr[1].value, NULL);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a:b");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 1);
+	ck_assert_ptr_eq(ptr, &buf[3]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[2]);
+	ck_assert_ptr_eq(hdr[1].name, NULL);
+	ck_assert_ptr_eq(hdr[1].value, NULL);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a:b\r\n");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 1);
+	ck_assert_ptr_eq(ptr, &buf[5]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[2]);
+	ck_assert_ptr_eq(hdr[1].name, NULL);
+	ck_assert_ptr_eq(hdr[1].value, NULL);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a:b\r\n\r\n");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 1);
+	ck_assert_ptr_eq(ptr, &buf[5]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[2]);
+	ck_assert_ptr_eq(hdr[1].name, NULL);
+	ck_assert_ptr_eq(hdr[1].value, NULL);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a: b\r\n");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 1);
+	ck_assert_ptr_eq(ptr, &buf[6]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[3]);
+	ck_assert_ptr_eq(hdr[1].name, NULL);
+	ck_assert_ptr_eq(hdr[1].value, NULL);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a :b\r\n");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 1);
+	ck_assert_ptr_eq(ptr, &buf[6]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[3]);
+	ck_assert_ptr_eq(hdr[1].name, NULL);
+	ck_assert_ptr_eq(hdr[1].value, NULL);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a : b\r\n");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 1);
+	ck_assert_ptr_eq(ptr, &buf[7]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[4]);
+	ck_assert_ptr_eq(hdr[1].name, NULL);
+	ck_assert_ptr_eq(hdr[1].value, NULL);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a: b\r\nc: d\r\n");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 2);
+	ck_assert_ptr_eq(ptr, &buf[12]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_str_eq(hdr[1].name, "c");
+	ck_assert_str_eq(hdr[1].value, "d");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[3]);
+	ck_assert_ptr_eq(hdr[1].name, &buf[6]);
+	ck_assert_ptr_eq(hdr[1].value, &buf[9]);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a: b\r\nc: d\r\n\r\n");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 2);
+	ck_assert_ptr_eq(ptr, &buf[12]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_str_eq(hdr[1].name, "c");
+	ck_assert_str_eq(hdr[1].value, "d");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[3]);
+	ck_assert_ptr_eq(hdr[1].name, &buf[6]);
+	ck_assert_ptr_eq(hdr[1].value, &buf[9]);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a: b\r\nc: d\r\n\r\ne: f\r\n");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 2);
+	ck_assert_ptr_eq(ptr, &buf[12]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_str_eq(hdr[1].name, "c");
+	ck_assert_str_eq(hdr[1].value, "d");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[3]);
+	ck_assert_ptr_eq(hdr[1].name, &buf[6]);
+	ck_assert_ptr_eq(hdr[1].value, &buf[9]);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a: b\r\nc: d");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 2);
+	ck_assert_ptr_eq(ptr, &buf[10]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_str_eq(hdr[1].name, "c");
+	ck_assert_str_eq(hdr[1].value, "d");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[3]);
+	ck_assert_ptr_eq(hdr[1].name, &buf[6]);
+	ck_assert_ptr_eq(hdr[1].value, &buf[9]);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a: b\r\nc: d\r\ne");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, -1);
+	/* The following could be undefined, since ret == -1 */
+	ck_assert_ptr_eq(ptr, &buf[12]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_str_eq(hdr[1].name, "c");
+	ck_assert_str_eq(hdr[1].value, "d");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[3]);
+	ck_assert_ptr_eq(hdr[1].name, &buf[6]);
+	ck_assert_ptr_eq(hdr[1].value, &buf[9]);
+	ck_assert_ptr_eq(hdr[2].name, NULL);
+	ck_assert_ptr_eq(hdr[2].value, NULL);
+
+	memset(hdr, 0, sizeof(hdr));
+	memset(buf, 0, sizeof(buf));
+	strcpy(buf, "a: b\r\nc: d\r\nefg:");
+	ptr = &buf[0];
+	ret = parse_http_headers(&ptr, hdr);
+	ck_assert_int_eq(ret, 3);
+	ck_assert_ptr_eq(ptr, &buf[16]);
+	ck_assert_str_eq(hdr[0].name, "a");
+	ck_assert_str_eq(hdr[0].value, "b");
+	ck_assert_str_eq(hdr[1].name, "c");
+	ck_assert_str_eq(hdr[1].value, "d");
+	ck_assert_str_eq(hdr[2].name, "efg");
+	ck_assert_str_eq(hdr[2].value, "");
+	ck_assert_ptr_eq(hdr[0].name, &buf[0]);
+	ck_assert_ptr_eq(hdr[0].value, &buf[3]);
+	ck_assert_ptr_eq(hdr[1].name, &buf[6]);
+	ck_assert_ptr_eq(hdr[1].value, &buf[9]);
+	ck_assert_ptr_eq(hdr[2].name, &buf[12]);
+	ck_assert_ptr_eq(hdr[2].value, &buf[16]);
+}
+END_TEST
+
+
 START_TEST(test_encode_decode)
 {
 	char buf[128];
@@ -719,67 +1352,8 @@ START_TEST(test_encode_decode)
 	int ret;
 	size_t len;
 
-#if defined(USE_WEBSOCKET) || defined(USE_LUA)
-	const char *alpha_b64_enc = "YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXo=";
-	const char *nonalpha_b64_enc =
-	    "ICEiIyQlJicoKSorLC0uLzAxMjM0NTY3ODk6Ozw9Pj9A";
-
 	mark_point();
 
-	memset(buf, 77, sizeof(buf));
-	base64_encode((unsigned char *)"a", 1, buf);
-	ck_assert_str_eq(buf, "YQ==");
-
-	memset(buf, 77, sizeof(buf));
-	base64_encode((unsigned char *)"ab", 1, buf);
-	ck_assert_str_eq(buf, "YQ==");
-
-	memset(buf, 77, sizeof(buf));
-	base64_encode((unsigned char *)"ab", 2, buf);
-	ck_assert_str_eq(buf, "YWI=");
-
-	memset(buf, 77, sizeof(buf));
-	base64_encode((unsigned char *)alpha, 3, buf);
-	ck_assert_str_eq(buf, "YWJj");
-
-	memset(buf, 77, sizeof(buf));
-	base64_encode((unsigned char *)alpha, 4, buf);
-	ck_assert_str_eq(buf, "YWJjZA==");
-
-	memset(buf, 77, sizeof(buf));
-	base64_encode((unsigned char *)alpha, 5, buf);
-	ck_assert_str_eq(buf, "YWJjZGU=");
-
-	memset(buf, 77, sizeof(buf));
-	base64_encode((unsigned char *)alpha, 6, buf);
-	ck_assert_str_eq(buf, "YWJjZGVm");
-
-	memset(buf, 77, sizeof(buf));
-	base64_encode((unsigned char *)alpha, (int)strlen(alpha), buf);
-	ck_assert_str_eq(buf, alpha_b64_enc);
-
-	memset(buf, 77, sizeof(buf));
-	base64_encode((unsigned char *)nonalpha, (int)strlen(nonalpha), buf);
-	ck_assert_str_eq(buf, nonalpha_b64_enc);
-#endif
-
-#if defined(USE_LUA)
-	memset(buf, 77, sizeof(buf));
-	len = 9999;
-	ret = base64_decode((unsigned char *)alpha_b64_enc,
-	                    (int)strlen(alpha_b64_enc),
-	                    buf,
-	                    &len);
-	ck_assert_int_eq(ret, -1);
-	ck_assert_uint_eq((unsigned int)len, (unsigned int)strlen(alpha));
-	ck_assert_str_eq(buf, alpha);
-
-	memset(buf, 77, sizeof(buf));
-	len = 9999;
-	ret = base64_decode((unsigned char *)"AAA*AAA", 7, buf, &len);
-	ck_assert_int_eq(ret, 3);
-#endif
-
 	memset(buf, 77, sizeof(buf));
 	ret = mg_url_encode(alpha, buf, sizeof(buf));
 	ck_assert_int_eq(ret, (int)strlen(buf));
@@ -1082,6 +1656,8 @@ START_TEST(test_config_options)
 #endif
 
 	ck_assert_str_eq("decode_url", config_options[DECODE_URL].name);
+	ck_assert_str_eq("decode_query_string",
+	                 config_options[DECODE_QUERY_STRING].name);
 
 #if defined(USE_LUA)
 	ck_assert_str_eq("lua_preload_file", config_options[LUA_PRELOAD_FILE].name);
@@ -1161,6 +1737,7 @@ make_private_suite(void)
 	TCase *const tcase_internal_parse_4 = tcase_create("Internal Parsing 4");
 	TCase *const tcase_internal_parse_5 = tcase_create("Internal Parsing 5");
 	TCase *const tcase_internal_parse_6 = tcase_create("Internal Parsing 6");
+	TCase *const tcase_internal_parse_7 = tcase_create("Internal Parsing 7");
 	TCase *const tcase_encode_decode = tcase_create("Encode Decode");
 	TCase *const tcase_mask_data = tcase_create("Mask Data");
 	TCase *const tcase_parse_date_string = tcase_create("Date Parsing");
@@ -1176,6 +1753,9 @@ make_private_suite(void)
 	suite_add_tcase(suite, tcase_http_keep_alive);
 
 	tcase_add_test(tcase_url_parsing_1, test_match_prefix);
+	tcase_add_test(tcase_url_parsing_1, test_match_prefix_strlen);
+	tcase_add_test(tcase_url_parsing_1, test_match_prefix_fuzz);
+	tcase_add_test(tcase_url_parsing_1, test_mg_match);
 	tcase_set_timeout(tcase_url_parsing_1, civetweb_min_test_timeout);
 	suite_add_tcase(suite, tcase_url_parsing_1);
 
@@ -1211,6 +1791,10 @@ make_private_suite(void)
 	tcase_set_timeout(tcase_internal_parse_6, civetweb_min_test_timeout);
 	suite_add_tcase(suite, tcase_internal_parse_6);
 
+	tcase_add_test(tcase_internal_parse_7, test_parse_http_headers);
+	tcase_set_timeout(tcase_internal_parse_7, civetweb_min_test_timeout);
+	suite_add_tcase(suite, tcase_internal_parse_7);
+
 	tcase_add_test(tcase_encode_decode, test_encode_decode);
 	tcase_set_timeout(tcase_encode_decode, civetweb_min_test_timeout);
 	suite_add_tcase(suite, tcase_encode_decode);
@@ -1254,6 +1838,7 @@ MAIN_PRIVATE(void)
 	test_parse_date_string(0);
 	test_parse_port_string(0);
 	test_parse_http_message(0);
+	test_parse_http_headers(0);
 	test_sha1(0);
 
 #if defined(_WIN32)
